import numpy as np
from scipy.spatial import KDTree
from scipy.optimize import brute


def generate_points(xx, yy, zz):
    """
    Function for creating a 3D-box of evenly distributed points.
    :param xx: the desired z dimensionality.
    :param yy: the desired y dimensionality.
    :param zz: the desired x dimensionality.
    :return: coordinates: a numpy array containing the 3D coordinates
    for the points.
    """
    coordinates = []
    for z in range(zz):
        for y in range(yy):
            for x in range(xx):
                coordinates.append("{} {} {}".format(x, y, z))
    coordinates = [w.split(' ') for w in coordinates]
    coordinates = np.asarray(coordinates).astype(np.float)
    coordinates = coordinates + ((np.asarray(np.random.rand(18, 3)) -
                                  0.5) * 0.05)
    coordinates = coordinates * [1, 1, 2]
    return coordinates


def generate_points_disk(number, radius, offset):
    """
    Function for generating randomly distributed 3D points along a disk.
    :param number: The desired number of points.
    :param radius: The radius of the disk.
    :param offset: The offset in the z direction.
    :return: A numpy array containing the 3D coordinates for the points.
    """
    r = np.random.uniform(low=0, high=radius, size=number)
    theta = np.random.uniform(low=0, high=2 * np.pi, size=number)
    noise = np.random.uniform(low=-0.05, high=0.05, size=number)

    x = np.sqrt(r) * np.cos(theta)
    y = np.sqrt(r) * np.sin(theta)
    z = ([offset] * number) + noise
    return np.asarray(list(zip(x, y, z)))


def generate_plane_svd(points):
    """
    Function for fitting a plane to a set of points by the Singular
    Value Decomposition method.
    :param points: A collection of points (x,y,z) that should be
    considered to find the best-fit plane.
    :return: x, y, z and normal: Returns the plane function z bounded
    by the max/min values of x and y structured in a meshgrid(xx, yy)
    for graphing. Returns the normal to the plane.
    """
    # defining the domain of the plane
    max_x = np.max(points[:, 0])
    max_y = np.max(points[:, 1])
    min_x = np.min(points[:, 0])
    min_y = np.min(points[:, 1])
    x, y = np.meshgrid([min_x, max_x], [min_y, max_y])
    # Note that the meshgrid is bounded by the x and y values,
    # meaning that for planes orthogonal to the xy-plane this will
    # produce a very bad visualization.

    # centering the data.
    centroid = points.mean(axis=0)
    points = points - centroid
    u, sigma, v = np.linalg.svd(points)
    normal = v[2]
    d = normal[0] * centroid[0] + normal[1] * centroid[1] + normal[2] \
        * centroid[2]

    z = (-normal[0] * x - normal[1] * y + d) * 1. / normal[2]
    return x, y, z, normal
    # https://stackoverflow.com/questions/53591350/plane-fit-of-3d-points-with-singular-value-decomposition
    # https://math.stackexchange.com/questions/2810048/plane-fitting-using-svd-normal-vector


def generate_plane_lstsq(points):
    """
    Function for fitting a plane to a set of points by the least
    squares method.
    :param points: A collection of points (x,y,z) that should be
    considered to find the best-fit plane.
    :return: x, y, z and normal: Returns the plane function z
    bounded  by the max/min values of x and y structured in a
    meshgrid(xx, yy) for graphing. Returns the normal to the plane.
    """
    matrix = np.column_stack((points[:, 0], points[:, 1],
                              np.ones(points[:, 0].size)))
    (a, b, c), residual, rank, s = \
        np.linalg.lstsq(matrix, points[:, 2], rcond=None)
    normal = unit_vector(np.asarray([a, b, -1]))
    # defining the domain of the plane
    max_x = np.max(points[:, 0])
    max_y = np.max(points[:, 1])
    min_x = np.min(points[:, 0])
    min_y = np.min(points[:, 1])
    # creating a point in the plane
    point = np.array([0.0, 0.0, c])
    d = -point.dot(normal)
    x, y = np.meshgrid([min_x, max_x], [min_y, max_y])
    z = (-normal[0] * x - normal[1] * y - d) * 1. / normal[2]
    # Comment regarding non-linear plane:
    # https://gist.github.com/amroamroamro/1db8d69b4b65e8bc66a6
    return x, y, z, normal


def unit_vector(vector):
    """
    Function for finding the unit vector of a vector.
    :param vector: The original vector.
    :return: The unit vector of the vector.
    """
    return vector / np.sqrt(vector.dot(vector))


def vector_3d(p1, p2):
    """
    Function for substracting one point/vector from another.
    :param p1: (x_a, y_a, z_a)
    :param p2: (x_b, y_b, z_b)
    :return: ((x_b - x_a), (y_b - y_a), (z_b - z_a))
    """
    return p2 - p1


def euclidean_distance(p1, p2):
    """
    Function for calculating the euclidean distance between two points.
    :param p1: Point A
    :param p2: Point B
    :return: The euclidean distance between point A and B.
    """
    vector = vector_3d(p1, p2)
    return np.sqrt(vector.dot(vector))


def angle_between_vectors(vector1, vector2):
    """
    Function for finding the angle between two vectors by applying
    the common formula : angle = arccos(vector1 dot vector2 /
    magnitude(vector1) * magnitude(vector2)).
    :param vector1: The first vector.
    :param vector2: The second vector.
    :return: The radian and angle between the two vectors.
    """
    v1 = unit_vector(vector1)
    v2 = unit_vector(vector2)

    radians = np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0))
    angle = radians * 180 / np.pi

    return radians, angle


def brute_optimal_angle(data_set, point, normal):
    """
    Function for determining the optimal angle based on brute force
    optimization.
    :param data_set: Data of (x,y,z) values / points to be evaluated.
    :param point: Point of origin.
    :param normal: The normal direction the points should be
    evaluated  against.
    :return: The point in data_set that lie closest in terms of
    angle  to the origin point.
    """
    def radians(n):
        v1 = unit_vector(vector_3d(point, data_set[int(n)]))
        v2 = unit_vector(normal)
        radian = min(np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0)),
        abs(abs(np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0))) - np.pi))
        return radian

    result = brute(radians, (slice(1, data_set.shape[0], 1),))
    return data_set[int(result[0])]


def brute_optimal_distance(data_set, point, normal):
    """
        Function for determining the optimal distance based on brute
        force optimization.
        :param data_set: Data of (x,y,z) values / points to be evaluated.
        :param point: Point of origin.
        :param normal: The normal direction the points should be
        evaluated against.
        :return: The point in data_set that lie closest in terms of
        distance to the normal vector.
        """
    def dist(n):
        cross_product = np.cross((data_set[int(n)] - point), normal)
        distance = np.sqrt(cross_product.dot(cross_product)) /  \
                   np.sqrt(normal.dot(normal))
        return distance

    result = brute(dist, (slice(1, data_set.shape[0], 1),))
    return data_set[int(result[0])]


def flatten_2d(data):
    """
    Function for reducing a numpy array down to two dimensions.
    :param data: A numpy array of n>2 dimensions.
    :return: A numpy array of n <= 2 dimensions.
    """
    return data.reshape(-1, data.shape[-1])


def pca(points):
    """
    Function for Principle Component construction for further analysis.
    :param points: An array of points to be evaluated.
    :return: The Eigenvalues and Eigenvectors of the data. The
    centered data, the mean and the covariance matrix.
    """
    mean = np.mean(points, axis=0)
    data = (points - mean)
    matrix = np.cov(data.T)
    eigenvalues, eigenvectors = np.linalg.eig(matrix)
    return eigenvalues, eigenvectors, data, mean, matrix


def kd_tree(data):
    """
    Function for creating a KD-tree for the given data.
    :param data: The data to be structured in a KD-tree structure.
    :return: The data structured in a KD-tree.
    """
    return KDTree(data)


def neighborhood_selection(tree, data_set, initial_point):
    """
    Function for finding the best neighbourhood point cluster selection.
    :param tree: KD-tree of the data
    :param data_set: The dataset to be evaluated, this is the same
    dataset as the KD-tree is structured by.
    :param initial_point: The initial point where the neighbourhood
    should be optimized around.
    :return: S_k, the optimal number of k nearest neighbours.
    """
    def shannon_entropy(points):
        epsilon = np.finfo(float).eps
        cov_matrix = np.cov(points.T)
        eigen_value, eigen_vector = np.linalg.eig(cov_matrix)
        idx = eigen_value.argsort()[::-1]
        eigen_value = eigen_value[idx]
        sum = eigen_value[0] + eigen_value[1] + eigen_value[2]
        eigen_value = eigen_value / sum
        linearity = (eigen_value[0] - eigen_value[1]) / (
                eigen_value[0]  + epsilon)
        planarity = (eigen_value[1] - eigen_value[2]) / (
                eigen_value[0]  + epsilon)
        scattering = eigen_value[2] / (eigen_value[0] + epsilon)
        shannonentropy = (- linearity * np.log(linearity + epsilon)
                          - planarity * np.log(planarity + epsilon)
                          - scattering * np.log(scattering + epsilon))
        return shannonentropy

    def query(n):
        distances, nearest_neighbours = \
            tree.query([data_set[initial_point]], k=int(n))
        return shannon_entropy(flatten_2d(data_set[nearest_neighbours]))

    result = brute(query, (slice(5+1, 18+1, 1),))
    # optimizes the minimum local solution within the given bounds
    # using brute force.
    # result = minimize_scalar(query, bounds=(5, 10), method='bounded')
    # optimizes the minimum local solution within the given bounds
    # using Bernt's Algorithm.
    return result[0]


def general_shannon_entropy(points):
    """
    Function for evaluating a collection of points (x,y,z) in terms
    of local features and entropy.
    :param points: A cluster of 3D coordinates (x,y,z)
    :return: The Covariance Matrix, Eigenvalues, Eigenvectors,
    Linearity, Planarity, Scattering, Omnivariance, Anisotropy,
    Curvature, Sum Eigenvalues and the Eigenentropy.
    """
    epsilon = np.finfo(float).eps
    cov_matrix = np.cov(points.T)
    eigen_value, eigen_vector = np.linalg.eig(cov_matrix)
    idx = eigen_value.argsort()[::-1]
    eigen_value = eigen_value[idx]
    eigen_vector = eigen_vector[:, idx]
    sum = eigen_value[0] + eigen_value[1] + eigen_value[2]
    eigen_value = eigen_value/sum
    linearity = (eigen_value[0] - eigen_value[1]) / (eigen_value[0]
                                                     + epsilon)
    planarity = (eigen_value[1] - eigen_value[2]) / (eigen_value[0]
                                                     + epsilon)
    scattering = eigen_value[2] / (eigen_value[0] + epsilon)
    omnivariance = abs(eigen_value[0] * eigen_value[1] *
                       eigen_value[2])**(1./3.)
    anisotropy = (eigen_value[0] - eigen_value[2]) / (eigen_value[0]
                                                      + epsilon)
    curvature = eigen_value[2]/(eigen_value[0] + eigen_value[1]
                                + eigen_value[2] + epsilon)
    sum_eigenvalues = eigen_value[0] + eigen_value[1] + eigen_value[2]
    eigenentropy = (- eigen_value[0] * np.log(eigen_value[0] + epsilon)
                    - eigen_value[1] * np.log(eigen_value[1] + epsilon)
                    - eigen_value[2] * np.log(eigen_value[2] + epsilon))
    shannonentropy = (  - linearity * np.log(linearity + epsilon)
                        - planarity * np.log(planarity + epsilon)
                        - scattering * np.log(scattering + epsilon))

    return cov_matrix, eigen_value, eigen_vector, linearity, \
           planarity, scattering, omnivariance, anisotropy,  \
           curvature, sum_eigenvalues, eigenentropy, shannonentropy


# def generate_plane(point, normal, plane_boundaries):
#     """
#     Function for creating a plane through a point orthogonal to the
#     normal vector.
#     :param point: The point the plane should lie in.
#     :param normal: The normal to the point and plane.
#     :param plane_boundaries: The plane's X-Y boundaries example:
#     X[0,5] Y[0,3]
#     :return x, y and z: Returns the plane function where
#     the boundaries are the mesh-grid defined by the boundaries of x
#     and y.
#     """
#     x, y = np.meshgrid(plane_boundaries, plane_boundaries)
#     x, y = np.meshgrid(plane_boundaries - (np.mean(x) - point[0]),
#                        plane_boundaries - (np.mean(y) - point[1]))
#     # d is the dot product between the point and the normal,
#     # this ensures the plane intersects the point.
#     d = -point.dot(normal)
#     # the formula for a plane is:           ax + by + cz + d = 0
#     # solving for the unknown z we have:    z = (-ax - by - d) / c
#     z = (- normal[0] * x - normal[1] * y - d) / normal[2]
#     return x, y, z

# def generate_plane_pca(points):
#     """
#     Function for fitting a plane to a set of points by the Principle Component Analysis method.
#     :param points: A collection of points (x,y,z) that should be considered to find the best-fit plane.
#     :return: x, y, z and normal: Returns the plane function z bounded by
#     the max/min values of x and y structured in a meshgrid(xx, yy) for graphing.
#     Returns the normal to the plane.
#     """
#     cov_matrix = np.cov(points.T)
#     eigen_value, eigen_vector = np.linalg.eig(cov_matrix)
#     idx = eigen_value.argsort()[::-1]
#     eigen_value = eigen_value[idx]
#     eigen_vector = eigen_vector[:, idx]
#     normal = eigen_vector[2]
#     d = -eigen_value.dot(normal)
#     # d = eigen_value[2]
#
#     # defining the domain of the plane
#     max_x = np.max(points[:, 0])
#     max_y = np.max(points[:, 1])
#     min_x = np.min(points[:, 0])
#     min_y = np.min(points[:, 1])
#
#     x, y = np.meshgrid([min_x, max_x], [min_y, max_y])
#     z = (-normal[0] * x - normal[1] * y - d) * 1. / normal[2]
#     return x, y, z, normal

# def import_points(path):
#     """
#     Function for importing and structuring raw data from a csv file.
#     :param path: The file path to a csv point cloud collection.
#     :return: The data structured as a numpy array.
#     """
#     with open(path, "r") as d:
#         data = d.readlines()
#     data = [w.translate({ord(';'): ","}).translate({ord('"'): None}).translate(
#         {ord('\n'): None}).replace('Point ', '').split(',') for w in data]
#     header = data[:1]
#     data = data[1:]
#     data = list(zip(*data))
#     data = np.asarray(data)
#     data = data[:, data[1].astype(np.int).argsort()]
#     data = list(zip(*data[2:5]))
#     data = np.asarray(data).astype(np.float)
#     return data